/* Modem.c
 * Michael McTernan, mm7323@bris.ac.uk
 * 3/10/00
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
  
#include "Modem.h"

#define DLE 0x010  
#define DTX 0x003  

#define S_DLE "\x010"
#define S_DTX "\x003"

/* Options that the user may control */
#ifdef DEBUG
int   debugFlags=3;
#else
int   debugFlags=0;
#endif
int   flowControl=2;
int   paritySetting=0;
int   dataBits=8;
int   stopBits=1;
char *modemDevice;

/* Some handy macros */
#define PARITY_NONE   (paritySetting==0)
#define PARITY_ODD    (paritySetting==1)
#define PARITY_EVEN   (paritySetting==2)
#define FLOW_XONXOFF  ((flowControl & 1)==1)
#define FLOW_CTSRTS   ((flowControl & 2)==2)
#define DBG_WRITES    ((debugFlags & 1)==1)
#define DBG_READS     ((debugFlags & 2)==2)


static int   modemStream;
static void  cleanExit();
static void  setModemLineSettings(int raw);


/*modemOpen*/
/* Opens the modem device 
 *  Returns true if all went well
 */
int modemOpen() {

  /* Check if there is a modem device set */
  if(!modemDevice) {
    modemDevice=strdup("/dev/modem");
    syslog(LOG_INFO,"Using default modem device %s",modemDevice);
  }

  syslog(LOG_INFO,"Opening modem device %s",modemDevice);

  modemStream=open(modemDevice,O_RDWR | O_SYNC);

  return modemStream>0;

}


/*modemInit*/
/* Setup the line options and modem device
 */
void modemInit() {

  setModemLineSettings(0);
      
  /* Do some initialisation */
  modemSendExpect("ATZ","OK");
  modemSendExpect("AT&K0","OK");        /* No flow control */
  if(FLOW_CTSRTS)
    modemSendExpect("AT&K3","OK");
  if(FLOW_XONXOFF)   
    modemSendExpect("AT&K4","OK");
    
  free(modemDevice);
  modemDevice=NULL;
}


/*modemSendExepect*/
/* Sends command to the modem, and
 *  returns if reply is on the first
 *  reply line.
 */
void modemSendExpect(const char *command,const char *reply) {
  modemWrite(command);
  modemExpect(command);
  modemExpect(reply);
}   


/*modemPoll*/
/* Returns 1 if there is data to read from
 *  the device.
 */
int modemPoll() {
  fd_set r,w,e;
  struct timeval timeout;
   
  timeout.tv_sec=0;
  timeout.tv_usec=0;

  FD_ZERO(&r);
  FD_ZERO(&w);
  FD_ZERO(&e);
    
  FD_SET(modemStream,&r);

  return select(modemStream+1,&r,&w,&e,&timeout)>0;

}


/*modemExpect*/
/* Reads line, and attempts to match String to
 *  a read line.  If there is no more input and
 *  String has not been matched, it fails.
 */
void modemExpect(const char *string) {
  char buf[1024];   

  do {

    /* Read a line */
    if(modemReadLine(buf,1024,1)<=0) {
      syslog(LOG_ERR,"Read from modem failed: %m");
      cleanExit();
    }

    /* Check the echo is correct */
    if(strcmp(string,buf)==0)
      return;

    syslog(LOG_ERR,"Unrecognised modem response");
    syslog(LOG_ERR,"  Expected: %s",string);
    syslog(LOG_ERR,"  Received: %s",buf);

  } while(modemPoll()>0);

  syslog(LOG_ERR,"  No more data to read.  Exiting.");
  cleanExit();
    
}


/*modemWrite*/
/* Writes *data to the modem device
 */
void modemWrite(const char *data) {
  int c=strlen(data);
    
  if(DBG_WRITES)
    syslog(LOG_INFO,"Writing: '%s'",data);
     
  while(c>0) {
    int w;
   
    w=write(modemStream,data,c);

    if(w<1) {
      syslog(LOG_ERR,"Write to modem device failed: %m");
      cleanExit();
    }
  
    c-=w;
    data+=w;

  } 
  
  if(write(modemStream,"\r",sizeof(char))!=sizeof(char)) {
    syslog(LOG_ERR,"Write to modem device failed: %m");
    cleanExit();
  }
   
}


/*modemRead*/
/* Reads raw data from the modem 
 */
int modemRead(char *buf,int n) {
  return read(modemStream,buf,n);
}


/*modemPlayFile*/
/* Plays a raw rockwell ADPCM file to the phone user.
 * If a DTMF char is detected then playing will
 *  cease, and the detected char will be returned.
 */
char modemPlayFile(const char *file,int enableDTMF) {
  int source,n;

  /* Config the modem for the correct sample rate and bits*/
  modemSendExpect("AT#VBS=4","OK");

  /* Switch the modem to voice transmit mode */
  modemSendExpect("AT#VTX","CONNECT");
  
  /* Turn off any translations in line settings */
  setModemLineSettings(1);
    
  /* Open the source file */
  if((source=open(file,O_RDONLY))<=0) {
    syslog(LOG_ERR,"open() failed for %s: %m",file);
    cleanExit();
  }

  /* Send all the data */
  do {
    char buf[1024];

    n=read(source,buf,1024);

    if(n>0) 
      printf("Writing: %d==%d\n",write(modemStream,buf,n),n);
    

  } while(n>0);

  /* Back to voice command mode */
  modemWrite(S_DLE S_DTX);
  modemExpect("VCON");

  /* Turn on any translations */
  setModemLineSettings(0);  

  return '\0';
}


/*modemReadDLE*/
/* Checks if any DLE commands have been
 *  returned from the modem, and returns
 *  the char value.  NULL otherwise
 * It uses to_sec and to_usec as the timeout.
 */ 
char modemReadDLE(long to_sec,long to_usec) {
  fd_set r,w,e;
  struct timeval timeout;
  char   c;

  timeout.tv_sec=to_sec;  
  timeout.tv_usec=to_usec;
  
  FD_ZERO(&r);
  FD_ZERO(&w);
  FD_ZERO(&e);

  FD_SET(modemStream,&r);
 
  /* Return if nothing is present to read */
  if(select(modemStream+1,&r,&w,&e,&timeout)==0)
    return '\0';

  /* Read the DLE char we are expecting */
  if(read(modemStream,&c,sizeof(char))<=0) {
    syslog(LOG_ERR,"read() failed: %m");
    cleanExit();
  }
  
  if(c!=DLE) {
    syslog(LOG_ERR,"Expected to read DLE, got 0x%x (%c)",c,c);
    cleanExit();
  }

  /* Read the actual code */
  if(read(modemStream,&c,sizeof(char))<=0) {
    syslog(LOG_ERR,"read() failed: %m");
    cleanExit();
  }

  return c;
}


/*modemReadLine*/
/* Reads a line from fd into the supplied
 *  buffer
 */
int modemReadLine(char *buf,int cbuf,int ignoreDLE) {
  int t;

  do {
    t=0;
    
    do {
      if(read(modemStream,&buf[t++],sizeof(char))<=0)
        return -1;

      if(ignoreDLE && buf[t-1]==DLE) {
        char c;

        if(read(modemStream,&c,sizeof(char))<=0)
          return -1;

        t--;
      }
 
    } while(buf[t-1]!='\n');
  
    buf[t-1]='\0';
    
    if(DBG_READS)
      syslog(LOG_INFO,"   Read: '%s'",buf);

  } while(t<=1);
      
  return t;

}


/*modemClose*/
/* Closes the modem device and
 *  hangs it up.
 */
void modemClose() {
  modemWrite(S_DLE S_DTX);
  modemWrite("ATH");
  close(modemStream);
}
     

/*cleanExit*/
/*
 */
static void cleanExit() {
  syslog(LOG_ERR,"Exiting");
  exit(-1);
}


/*setModemLineSettings*/
/* Sets up the char conversions etc...
 * raw allows all conversions to be disabled
 *  pehaps for voice IO 
 */
static void setModemLineSettings(int raw) {
  struct termios term;

  /* Get the line options */
  if(tcgetattr(modemStream,&term)!=0) {
    syslog(LOG_ERR,"tcgetattr failed: %m");
    cleanExit();
  }
  
  /* Input flags */
  if(raw==1) {
    term.c_iflag&=~(ISTRIP | INLCR | ICRNL | IUCLC | IGNCR);
  } else {
    term.c_iflag&=~(ISTRIP | INLCR | ICRNL | IUCLC);
    term.c_iflag|=IGNCR;
  }
  
  if(FLOW_XONXOFF)
    term.c_iflag|=IXON | IXOFF;
  else
    term.c_iflag&=~(IXON | IXOFF);

  /* Output flags */
  term.c_oflag&=~(OPOST | OLCUC | ONLCR | OCRNL);

  /* Control flags */
  term.c_cflag|=CREAD | CRTSCTS | HUPCL;   
  if(FLOW_CTSRTS)
    term.c_cflag|=CRTSCTS;
  else
    term.c_cflag&=~CRTSCTS;

  switch(dataBits) {  
  case 5:
    term.c_cflag|=CS5;
    break;
  case 6:
    term.c_cflag|=CS6;
    break;
  case 7:
    term.c_cflag|=CS7;
    break;
  case 8:
    term.c_cflag|=CS8;
    break;
  default:
    syslog(LOG_ERR,"Invalid number of character bits");
    cleanExit();
  }

  if(stopBits==2)
    term.c_cflag|=CSTOPB;
  else
    term.c_cflag&=~CSTOPB;
    
  if(PARITY_NONE)
    term.c_cflag&=~PARENB;
  else {
    if(PARITY_ODD)
      term.c_cflag|=~PARODD;
    else
      term.c_cflag&=~PARODD;
  }

  /* Local flags */
  term.c_lflag&=~(ECHO | ISIG | ICANON | IEXTEN | OFILL);

  /* Set the bauds */
  if(cfsetospeed(&term,B115200)!=0 || cfsetispeed(&term,B115200)!=0) {
    syslog(LOG_ERR,"cfsetospeed() or cfsetispeed() failed: %m");
    cleanExit();
  } 

  /* Write them back */   
  if(tcsetattr(modemStream,TCSAFLUSH,&term)!=0) {
    syslog(LOG_ERR,"tcgetattr failed: %m");
    cleanExit();
  }

}
